"""REST helpers for Muffin Framework."""

__version__ = "2.6.4"
__project__ = "muffin-rest"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"

# Default query params
FILTERS_PARAM = 'where'
LIMIT_PARAM = 'limit'
OFFSET_PARAM = 'offset'
SORT_PARAM = 'sort'


from .api import API                # noqa
from .handler import RESTHandler    # noqa
from .errors import APIError        # noqa

# Just an alias to support legacy style
Api = API


try:
    from .peewee import PWRESTHandler, PWFilter
except ImportError:
    pass


try:
    from .sqlalchemy import SARESTHandler, SAFilter
except ImportError:
    pass


try:
    from .mongo import MongoRESTHandler, MongoFilter
except ImportError:
    pass



"""Support openapi."""

import typing as t

from apispec import APISpec
from http_router.routes import Route

from ..openapi import OpenAPIMixin


class PeeweeOpenAPIMixin(OpenAPIMixin):
    """Render openapi."""

    if t.TYPE_CHECKING:
        from . import PWRESTOptions

        meta: PWRESTOptions

    @classmethod
    def openapi(cls, route: Route, spec: APISpec) -> t.Dict:
        """Get openapi specs for the endpoint."""
        operations = super(PeeweeOpenAPIMixin, cls).openapi(route, spec)
        is_resource_route = getattr(route, 'params', {}).get(cls.meta.name_id)
        if not is_resource_route and 'delete' in operations:
            operations['delete'].setdefault('parameters', [])
            operations['delete']['requestBody'] = {
                'required': True,
                'content': {
                    'application/json': {'schema': {'type': 'array', 'items': {'type': 'string'}}}
                }
            }
        return operations



"""Create openapi schema from the given API."""
import typing as t
import inspect
import re
from http import HTTPStatus
from functools import partial

from apispec import APISpec, utils
from apispec.ext.marshmallow import MarshmallowPlugin
from http_router.routes import DynamicRoute, Route
from asgi_tools.response import CAST_RESPONSE
from muffin import Response
from muffin.typing import JSONType

from . import FILTERS_PARAM, LIMIT_PARAM, OFFSET_PARAM, SORT_PARAM, openapi

try:
    from apispec import yaml_utils
except ImportError:
    yaml_utils = None


DEFAULT_METHODS = 'get',
HTTP_METHODS = ['GET', 'POST', 'PUT', 'PATH', 'DELETE', 'HEAD', 'OPTIONS', 'TRACE', 'CONNECT']
RE_URL = re.compile(r'<(?:[^:<>]+:)?([^<>]+)>')
SKIP_PATH = {'/openapi.json', '/swagger', '/redoc'}


def render_openapi(api, request):
    """Prepare openapi specs."""
    # Setup Specs
    options = dict(api.openapi_options)
    options.setdefault('servers', [{
        'url': str(request.url.with_query('').with_path(api.prefix))
    }])

    spec = APISpec(
        options['info'].pop('title', f"{ api.app.cfg.name.title() } API"),
        options['info'].pop('version', '1.0.0'),
        options.pop('openapi_version', '3.0.0'),
        **options, plugins=[MarshmallowPlugin()])
    spec.tags = {}

    # Setup Authorization
    if api.authorize:
        _, _, schema = parse_docs(api.authorize)
        spec.options['security'] = []
        for key, value in schema.items():
            spec.components.security_scheme(key, value)
            spec.options['security'].append({key: []})

    # Setup Paths
    routes = api.router.routes()
    for route in routes:
        if route.path in SKIP_PATH:
            continue

        spec.path(route.path, **route_to_spec(route, spec))

    return spec.to_dict()


def parse_docs(cb: t.Callable) -> t.Tuple[str, str, t.Dict]:
    """Parse docs from the given callback."""
    if yaml_utils is None:
        return '', '', {}

    docs = cb.__doc__ or ''
    schema = yaml_utils.load_yaml_from_docstring(docs)
    docs = docs.split('---')[0]
    docs = utils.dedent(utils.trim_docstring(docs))
    summary, _, description = docs.partition('\n\n')
    return summary, description.strip(), schema


def merge_dicts(source: t.Dict, merge: t.Dict) -> t.Dict:
    """Merge dicts."""
    return dict(source, **{
        key: ((
            merge_dicts(source[key], merge[key])
            if isinstance(source[key], dict) and isinstance(merge[key], dict)
            else (
                source[key] + merge[key]
                if isinstance(source[key], list) and isinstance(merge[key], list)
                else merge[key]
            )
        ) if key in source else merge[key]) for key in merge})


def route_to_spec(route: Route, spec: APISpec) -> t.Dict:
    """Convert the given router to openapi operations."""
    results: t.Dict = {'parameters': [], 'operations': {}}
    if isinstance(route, DynamicRoute):
        for param in route.params:
            results['parameters'].append({'in': 'path', 'name': param})

    target = t.cast(t.Callable, route.target)
    if isinstance(target, partial):
        target = target.func

    if hasattr(target, 'openapi'):
        results['operations'] = target.openapi(route, spec) # type: ignore
        return results

    summary, desc, schema = parse_docs(target)
    responses = return_type_to_response(target)
    for method in route_to_methods(route):
        results['operations'][method] = {
            'summary': summary,
            'description': desc,
            'responses': responses
        }

    results['operations'] = merge_dicts(results['operations'], schema)
    return results


def route_to_methods(route: Route) -> t.List[str]:
    """Get sorted methods from the route."""
    methods = [m for m in HTTP_METHODS if m in (route.methods or [])]
    return [m.lower() for m in methods or DEFAULT_METHODS]


def return_type_to_response(fn: t.Callable) -> t.Dict:
    """Generate reponses specs based on the given function's return type."""
    responses: t.Dict[int, t.Dict] = {}
    return_type = fn.__annotations__.get('return')
    return_type = CAST_RESPONSE.get(return_type, return_type)  # type: ignore
    if return_type is None:
        return responses

    if inspect.isclass(return_type) and issubclass(return_type, Response) and \
            return_type.content_type:

        responses[return_type.status_code] = {
            'description': HTTPStatus(return_type.status_code).description,
            'content': {
                return_type.content_type: {
                }
            }
        }
    return responses


class OpenAPIMixin:
    """Render an endpoint to openapi specs."""

    if t.TYPE_CHECKING:
        from .endpoint import RESTOptions

        meta: RESTOptions

    @classmethod
    def openapi(cls, route: Route, spec: APISpec) -> t.Dict:
        """Get openapi specs for the endpoint."""
        operations: t.Dict = {}
        summary, desc, schema = parse_docs(cls)
        if cls not in spec.tags:
            spec.tags[cls] = cls.meta.name
            spec.tag({'name': cls.meta.name, 'description': summary})
            spec.components.schema(cls.meta.Schema.__name__, schema=cls.meta.Schema)

        schema_ref = {'$ref': f"#/components/schemas/{ cls.meta.Schema.__name__ }"}
        for method in route_to_methods(route):
            operations[method] = {'tags': [spec.tags[cls]]}
            is_resource_route = isinstance(route, DynamicRoute) and \
                route.params.get(cls.meta.name_id)

            if method == 'get' and not is_resource_route:
                operations[method]['parameters'] = []
                if cls.meta.sorting:
                    sorting = list(cls.meta.sorting.keys())
                    operations[method]['parameters'].append({
                        'name': SORT_PARAM, 'in': 'query', 'style': 'form', 'explode': False,
                        'schema': {'type': 'array', 'items': {'type': 'string', 'enum': sorting}},
                        'description': ",".join(sorting),
                    })

                if cls.meta.filters.filters:
                    operations[method]['parameters'].append({
                        'name': FILTERS_PARAM, 'in': 'query', 'description': str(cls.meta.filters),
                        'content': {'application/json': {'schema': {'type': 'object'}}}
                    })

                if cls.meta.limit:
                    operations[method]['parameters'].append({
                        'name': LIMIT_PARAM, 'in': 'query',
                        'schema': {'type': 'integer', 'minimum': 1, 'maximum': cls.meta.limit},
                        'description': 'The number of items to return',
                    })
                    operations[method]['parameters'].append({
                        'name': OFFSET_PARAM, 'in': 'query',
                        'schema': {'type': 'integer', 'minimum': 0},
                        'description': 'The offset of items to return',
                    })

            # Update from the method
            meth = getattr(cls, method, None)
            if isinstance(route.target, partial) and '__meth__' in route.target.keywords:
                meth = getattr(cls, route.target.keywords['__meth__'], None)

            elif method in {'post', 'put'}:
                operations[method]['requestBody'] = {
                    'required': True, 'content': {'application/json': {'schema': schema_ref}}
                }

            if meth:
                operations[method]['summary'], operations[method]['description'], mschema = openapi.parse_docs(meth)  # noqa
                return_type = meth.__annotations__.get('return')
                if return_type == 'JSONType' or return_type == JSONType:
                    responses = {200: {'description': 'Request is successfull', 'content': {
                        'application/json': {'schema': schema_ref}
                    }}}
                else:
                    responses = return_type_to_response(meth)

                operations[method]['responses'] = responses
                operations[method] = merge_dicts(operations[method], mschema)

        return merge_dicts(operations, schema)



"""Support API filters."""
import operator
import typing as t

import marshmallow as ma


class Filter:

    """Base filter class."""

    operators = {
        '<':   operator.lt,
        '$lt': operator.lt,
        '<=':  operator.le,
        '$le': operator.le,
        '>':   operator.gt,
        '$gt': operator.gt,
        '>=':  operator.ge,
        '$ge': operator.ge,
        '==':  operator.eq,
        '$eq': operator.eq,
        '!':   operator.ne,
        '$ne': operator.ne,
        '$in': lambda v, c: v in c,
        '$nin': lambda v, c: v not in c,
    }
    operators['<<'] = operators['$in']

    list_ops = ['$in', '<<']

    field_cls: t.Type[ma.fields.Field] = ma.fields.Raw

    def __init__(self, name: str, attr: str = None, field: ma.fields.Field = None):
        """Initialize filter.

        :param name: The filter's name
        :param attr: Column/Property name.
        :param fields: Marshmallow Field instance

        """
        self.name = name
        self.attr = attr or name
        self.field = field or self.field_cls(attribute=self.attr)

    def __repr__(self) -> str:
        """Represent self as a string."""
        return '<Filter %s>' % self.name

    def filter(self, collection: t.Any, data: t.Mapping, **kwargs):
        """Filter given collection."""
        try:
            ops = self.parse(data)
        except ma.ValidationError:
            return None, collection

        collection = self.apply(collection, *ops, **kwargs)
        return ops, collection

    def parse(self, data: t.Mapping) -> t.Tuple[t.Tuple[t.Callable, t.Any], ...]:
        """Parse operator and value from filter's data."""
        val = data.get(self.name, ma.missing)
        if not isinstance(val, dict):
            return (self.operators['$eq'], self.field.deserialize(val)),

        return tuple(
            (
                self.operators[op],
                (self.field.deserialize(val)) if op not in self.list_ops else [
                    self.field.deserialize(v) for v in val])
            for (op, val) in val.items() if op in self.operators
        )

    def apply(self, collection, *ops: t.Tuple[t.Callable, t.Any], **kwargs):
        """Apply the filter to collection."""
        validator = lambda obj: all(op(get_value(obj, self.name), val) for (op, val) in ops)  # noqa
        return [o for o in collection if validator(o)]


class Filters:

    """Build filters for given handler."""

    FILTER_CLASS = Filter

    def __init__(self, *filters, handler=None):
        """Initialize object."""
        self.filters = tuple(
            f if isinstance(f, Filter) else self.convert(f, handler) for f in filters)

    def __iter__(self):
        """Iterate through self filters."""
        return iter(self.filters)

    def __str__(self) -> str:
        """Describe the filters."""
        return ", ".join(f.name for f in self)

    def convert(self, args, handler=None):
        """Prepare filters."""
        name = args
        field = attr = None
        opts = ()
        if isinstance(args, (list, tuple)):
            name, *opts = args
            if opts:
                attr = opts.pop()
            if opts:
                field = opts.pop()

        if not field and handler and handler.meta.Schema:
            field = handler.meta.Schema._declared_fields.get(attr or name) or \
                self.FILTER_CLASS.field_cls()
            field.attribute = field.attribute or attr or name
        return self.FILTER_CLASS(name, attr=attr, field=field, *opts)

    def filter(self, data, collection, **kwargs):
        """Filter given collection."""
        if not data or self.filters is None:
            return None, collection

        filters = {}
        for f in self.filters:
            if f.name not in data:
                continue
            ops, collection = f.filter(collection, data, **kwargs)
            filters[f.name] = ops

        return filters, collection


def get_value(obj, name):
    """Get value from object by name."""
    if isinstance(obj, dict):
        return obj.get(name)

    return getattr(obj, name, obj)



"""Mongo Utils."""

from __future__ import annotations  # <py3.10

import typing as t

from motor import motor_asyncio as motor


class MongoChain:

    """Support query chains.

    Only for `find` and `find_one` methods.
    ::
        collection = MongoChain(mongo_collection)
        collection = collection.find({'field': 'value').find('field2': 'value')
        result = collection.find_one({'field3': 'value')
        results = collection.skip(10).limit(10)
    """

    CURSOR_METHODS = (
        'where', 'sort', 'skip', 'rewind', 'retrieved', 'remove_option', 'next', 'min',
        'max_time_ms', 'max_scan', 'max_await_time_ms', 'max', 'limit', 'hint', 'explain',
        'distinct', 'cursor_id', 'count', 'comment', 'collection', 'close', 'clone', 'batch_size',
        'alive', 'address', 'add_option', '__getitem__'
    )

    def __init__(self, collection: motor.AsyncIOMotorCollection):
        """Initialize the resource."""
        self.collection = collection
        self.query: t.List = []
        self.projection = None
        self.sorting = None

    def find(self, query: t.Union[t.List, t.Dict] = None, projection=None) -> MongoChain:
        """Store filters in self."""
        self.query = self.__update__(query)
        self.projection = projection
        return self

    def find_one(self, query: t.Union[t.List, t.Dict] = None, projection=None) -> t.Awaitable:
        """Apply filters and return cursor."""
        query = self.__update__(query)
        query = query and {'$and': query} or {}
        return self.collection.find_one(query, projection=projection)

    def count(self) -> int:
        query = self.query and {'$and': self.query} or {}
        return self.collection.count_documents(query)

    def aggregate(self, pipeline, **kwargs):
        """Aggregate collection."""
        if self.query:
            for params in pipeline:
                if '$match' in params:
                    query = self.__update__(params['$match'])
                    params['$match'] = {'$and': query}
                    break
            else:
                pipeline.insert(0, {'$match': {'$and': query}})

        if self.sorting:
            pipeline = [p for p in pipeline if '$sort' not in p]
            pipeline.append({'$sort': dict(self.sorting)})

        return self.collection.aggregate(pipeline, **kwargs)

    def sort(self, key, direction=1):
        """Save ordering properties."""
        if isinstance(key, str):
            self.sorting = [(key, direction)]
        else:
            self.sorting = key

        return self

    def __repr__(self):
        """String representation."""
        return "<MongoChain (%s) %r>" % (self.collection.name, self.query)

    def __update__(self, query):
        """Update stored query."""
        if query:
            self.query.append(query)

        return self.query

    def __iter__(self):
        """Iterate by self collection."""
        query = self.query and {'$and': self.query} or {}
        if self.sorting:
            return self.collection.find(query, self.projection).sort(self.sorting)

        return self.collection.find(query, self.projection)

    def __getattr__(self, name):
        """Proxy any attributes except find to self.collection."""
        if name in self.CURSOR_METHODS:
            query = self.query and {'$and': self.query} or {}
            cursor = self.collection.find(query, self.projection)
            if self.sorting:
                cursor = cursor.sort(self.sorting)
            return getattr(cursor, name)
        return getattr(self.collection, name)



"""SQLAlchemy Core support."""

from __future__ import annotations  # py37

import typing as t

import marshmallow as ma
import muffin
import sqlalchemy as sa
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema as BaseSQLAlchemyAutoSchema, ModelConverter
from muffin.typing import JSONType
from muffin_databases import Plugin as DB

from ..handler import RESTHandler, RESTOptions
from ..errors import APIError
from ..filters import Filter, Filters


# XXX: Monkey patch ModelConverter
ModelConverter._get_field_name = lambda s, prop_or_column: str(prop_or_column.key)


class SQLAlchemyAutoSchema(BaseSQLAlchemyAutoSchema):
    """Allow partial updates for tables."""

    @ma.pre_load
    def fill_defaults(self, data, partial=False, **kwargs):
        """Insert default params for SQLAlchemy because databases don't.

        https://github.com/encode/databases/issues/72
        """
        cols_to_fields = {
            f.attribute or f.name: f for f in self.declared_fields.values()}
        if not partial:
            for column in self.opts.table.columns:
                field = cols_to_fields.get(column.name)
                if not field:
                    continue

                data_key = field.data_key or field.name
                if data_key not in data and column.default is not None:
                    value = column.default.arg
                    if callable(value):
                        value = value(column)

                    data[data_key] = field._serialize(value, field.attribute or field.name, None)

        return data

    @ma.post_load
    def make_instance(self, data, **kwargs):
        """Update a table instance."""
        if self.instance is not None:
            instance = dict(self.instance)
            instance.update(data)
            return instance
        return super(SQLAlchemyAutoSchema, self).make_instance(data, **kwargs)


class SAFilter(Filter):
    """Custom filter for sqlalchemy."""

    operators = Filter.operators
    operators['$between'] = lambda c, v: c.between(*v)
    operators['$ends'] = lambda c, v: c.endswith(v)
    operators['$ilike'] = lambda c, v: c.ilike(v)
    operators['$in'] = lambda c, v: c.in_(v)
    operators['$like'] = lambda c, v: c.like(v)
    operators['$match'] = lambda c, v: c.match(v)
    operators['$nin'] = lambda c, v: ~c.in_(v)
    operators['$notilike'] = lambda c, v: c.notilike(v)
    operators['$notlike'] = lambda c, v: c.notlike(v)
    operators['$starts'] = lambda c, v: c.startswith(v)

    list_ops = Filter.list_ops + ['$between']

    def __init__(self, name: str, attr: str = None,
                 field: ma.fields.Field = None, column: sa.Column = None):
        """Support custom model fields."""
        super().__init__(name, attr, field)
        self.column = column

    def apply(self, collection: sa.sql.Select, *ops: t.Tuple[t.Callable, t.Any],
              handler: SARESTHandler = None, **kwargs) -> sa.sql.Select:
        """Apply the filters to SQLAlchemy Select."""
        column = self.column
        if column is None and handler:
            column = handler.meta.table.c.get(self.field.attribute or self.field.name)

        if ops and column is not None:
            return self.query(collection, column, *ops, **kwargs)

        return collection

    def query(self, select: sa.sql.Select, column: sa.Column, *ops, **kwargs) -> sa.sql.Select:
        """Filter a select."""
        return select.where(*[op(column, val) for op, val in ops])


class SAFilters(Filters):
    """Bind SAfilter class."""

    FILTER_CLASS = SAFilter


class SARESTOptions(RESTOptions):
    """Support SQLAlchemy Core."""

    # Base filters class
    filters_cls: t.Type[SAFilters] = SAFilters

    # Schema auto generation params
    schema_base: t.Type[SQLAlchemyAutoSchema] = SQLAlchemyAutoSchema

    if t.TYPE_CHECKING:
        table: sa.Table
        table_pk: sa.Column
        database: DB
        Schema: t.Type[SQLAlchemyAutoSchema]

    def setup(self, cls):
        """Prepare meta options."""
        if self.table is None:
            raise ValueError("'SARESTHandler.Meta.table' is required")

        if self.database is None:
            raise ValueError("'SARESTHandler.Meta.database' is required")

        self.table_pk = self.table_pk or self.table.c.id

        self.name = self.name or self.table.name
        self.name_id = self.name_id or self.table_pk.name

        super(SARESTOptions, self).setup(cls)

        # Setup sorting
        self.sorting = dict(
            (isinstance(n, sa.Column) and n.name or n, prop)
            for (n, prop) in self.sorting.items())

    def setup_schema_meta(self, cls):
        """Prepare a schema."""
        return type('Meta', (object,), dict({
            'unknown': self.schema_unknown, 'table': self.table,
            'include_fk': True, 'dump_only': (self.name_id,)
        }, **self.schema_meta))


class SARESTHandler(RESTHandler):
    """Support SQLAlchemy Core."""

    meta: SARESTOptions
    meta_class: t.Type[SARESTOptions] = SARESTOptions

    class Meta:
        """Tune the handler."""

        abc = True

        # Sqlalchemy options
        table = None
        table_pk = None
        database = None

    async def prepare_collection(self, request: muffin.Request) -> sa.sql.Select:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return self.meta.table.select()

    async def paginate(self, request: muffin.Request, *, limit: int = 0,
                       offset: int = 0) -> t.Tuple[sa.sql.Select, int]:
        """Paginate the collection."""
        qs = sa.select([sa.func.count()]).select_from(self.collection.order_by(None))
        total = await self.meta.database.fetch_val(qs)
        return self.collection.offset(offset).limit(limit), total

    async def get(self, request, *, resource=None) -> JSONType:
        """Get resource or collection of resources."""
        if resource is not None and resource != '':
            return await self.dump(request, resource, many=False)

        rows = await self.meta.database.fetch_all(self.collection)
        return await self.dump(request, rows, many=True)

    async def prepare_resource(self, request: muffin.Request) -> t.Optional[dict]:
        """Load a resource."""
        pk = request['path_params'].get(self.meta.name_id)
        if not pk:
            return None

        qs = self.collection.where(self.meta.table_pk == pk)
        resource = await self.meta.database.fetch_one(qs)
        if resource is None:
            raise APIError.NOT_FOUND('Resource not found')
        return dict(resource)

    async def get_schema(self, request: muffin.Request, resource=None) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get('schema_only'),
            exclude=request.url.query.get('schema_exclude', ()),
        )

    async def save(self, request: muffin.Request,  # type: ignore
                   resource: t.Union[dict, t.List[dict]]):
        """Save the given resource.

        Supports batch saving.
        """
        resources = resource if isinstance(resource, t.Sequence) else [resource]
        for res in resources:
            if res.get(self.meta.table_pk.name):
                update = self.meta.table.update().where(
                    self.meta.table_pk == res[self.meta.table_pk.name])
                await self.meta.database.execute(update, res)

            else:
                insert = self.meta.table.insert()
                res[self.meta.table_pk.name] = await self.meta.database.execute(insert, res)

        return resource

    async def remove(self, request: muffin.Request, *, resource: dict = None):
        """Remove the given resource."""
        pks = [resource[self.meta.table_pk.name]] if resource else await request.data()
        if not pks:
            raise APIError.NOT_FOUND()

        delete = self.meta.table.delete(self.meta.table_pk.in_(pks))
        await self.meta.database.execute(delete)

    delete = remove  # noqa

    async def sort(self, request: muffin.Request,
                   *sorting: t.Tuple[str, bool], **options) -> sa.sql.Select:
        """Sort the current collection."""
        order_by = []
        for name, desc in sorting:
            column = self.meta.table.c.get(name)
            if column is None:
                continue

            if desc:
                column = column.desc()  # type: ignore

            order_by.append(column)

        if order_by:
            return self.collection.order_by(*order_by)

        return self.collection



"""Support for Peewee ORM (https://github.com/coleifer/peewee)."""

from __future__ import annotations

import typing as t

import marshmallow as ma
import muffin
import peewee as pw
from apispec.ext.marshmallow import MarshmallowPlugin
from marshmallow_peewee import ModelSchema, ForeignKey

from ..handler import RESTBase, RESTOptions
from ..errors import APIError
from ..filters import Filter, Filters

from .openapi import PeeweeOpenAPIMixin


# XXX: Patch apispec.MarshmallowPlugin to support ForeignKeyField
MarshmallowPlugin.Converter.field_mapping[ForeignKey] = ("integer", None)


class PWFilter(Filter):
    """Support Peewee."""

    operators = Filter.operators
    operators['$in'] = lambda f, v: f << v
    operators['$none'] = lambda f, v: f >> v
    operators['$like'] = lambda f, v: f % v
    operators['$contains'] = lambda f, v: f.contains(v)
    operators['$starts'] = lambda f, v: f.startswith(v)
    operators['$ends'] = lambda f, v: f.endswith(v)
    operators['$between'] = lambda f, v: f.between(*v)
    operators['$regexp'] = lambda f, v: f.regexp(v)

    list_ops = Filter.list_ops + ['$between']

    def __init__(self, name: str, attr: str = None,
                 field: ma.fields.Field = None, pw_field: pw.Field = None):
        """Support custom model fields."""
        super(PWFilter, self).__init__(name, attr, field)
        self.pw_field = pw_field

    def apply(self, collection: pw.Query, *ops: t.Tuple[t.Callable, t.Any],
              handler: PWRESTBase = None, **kwargs) -> pw.Query:
        """Apply the filters to Peewee QuerySet.."""
        pw_field = (
            self.pw_field or handler and handler.meta.model._meta.fields.get(self.field.attribute))
        if pw_field and ops:
            return self.query(collection, pw_field, *ops, **kwargs)
        return collection

    def query(self, query: pw.Query, column: pw.Field, *ops: t.Tuple, **kwargs) -> pw.Query:
        """Filter a query."""
        return query.where(*[op(column, val) for op, val in ops])


class PWFilters(Filters):
    """Bind Peewee filter class."""

    FILTER_CLASS = PWFilter


class PWRESTOptions(RESTOptions):
    """Support Peewee."""

    # Base filters class
    filters_cls: t.Type[PWFilters] = PWFilters

    # Schema auto generation params
    schema_base: t.Type[ModelSchema] = ModelSchema

    if t.TYPE_CHECKING:
        model: pw.Model
        model_pk: pw.Field
        Schema: t.Type[ModelSchema]

    def setup(self, cls):
        """Prepare meta options."""
        if not self.model:
            raise ValueError("'PWRESTHandler.Meta.model' is required")

        self.model_pk = self.model_pk or self.model._meta.primary_key

        self.name = self.name or self.model._meta.table_name
        self.name_id = self.name_id or self.model_pk.name

        super(PWRESTOptions, self).setup(cls)

        # Setup sorting
        self.sorting = dict(
            (isinstance(n, pw.Field) and n.name or n, prop)
            for (n, prop) in self.sorting.items())

    def setup_schema_meta(self, cls):
        """Prepare a schema."""
        return type('Meta', (object,), dict(
            {'unknown': self.schema_unknown, 'model': self.model}, **self.schema_meta))


class PWRESTBase(RESTBase):
    """Support Peeweee."""

    collection: pw.Query
    resource: pw.Model
    meta: PWRESTOptions
    meta_class: t.Type[PWRESTOptions] = PWRESTOptions

    class Meta:
        """Tune the handler."""

        abc: bool = True

        # Peewee options
        model = None
        model_pk = None

    async def prepare_collection(self, request: muffin.Request) -> pw.Query:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return self.meta.model.select()

    async def prepare_resource(self, request: muffin.Request) -> t.Optional[pw.Model]:
        """Load a resource."""
        pk = request['path_params'].get(self.meta.name_id)
        if not pk:
            return None

        try:
            return self.collection.where(self.meta.model_pk == pk).get()
        except self.meta.model.DoesNotExist:
            raise APIError.NOT_FOUND('Resource not found')

    async def sort(self, request: muffin.Request,
                   *sorting: t.Tuple[str, bool], **options) -> pw.Query:
        """Sort the current collection."""
        order_by = []
        for name, desc in sorting:
            field = self.meta.sorting.get(name)
            if field and not isinstance(field, pw.Field):
                field = self.meta.model._meta.fields.get(name)

            if field is None:
                continue

            if desc:
                field = field.desc()  # type: ignore

            order_by.append(field)

        if order_by:
            return self.collection.order_by(*order_by)

        return self.collection

    async def paginate(self, request: muffin.Request, *, limit: int = 0,
                       offset: int = 0) -> t.Tuple[pw.Query, int]:
        """Paginate the collection."""
        cqs = self.collection.order_by()
        if cqs._group_by:
            cqs._select = cqs._group_by
        return self.collection.offset(offset).limit(limit), cqs.count()

    async def save(self, request: muffin.Request,  # type: ignore
                   resource: t.Union[pw.Model, t.List[pw.Model]]):
        """Save the given resource.

        Supports batch saving.
        """
        for obj in (resource if isinstance(resource, list) else [resource]):
            obj.save()
        return resource

    async def remove(self, request: muffin.Request, *, resource: pw.Model = None):
        """Remove the given resource."""
        if resource:
            resources = [resource]

        else:
            data = await request.data()
            if not data:
                return
            resources = list(self.collection.where(self.meta.model_pk << data))

        if not resources:
            raise APIError.NOT_FOUND()

        for resource in resources:
            resource.delete_instance()

    delete = remove  # noqa

    async def get_schema(self, request: muffin.Request, resource=None) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get('schema_only'),
            exclude=request.url.query.get('schema_exclude', ()),
        ) if self.meta.Schema else None


class PWRESTHandler(PWRESTBase, PeeweeOpenAPIMixin):  # type: ignore
    """Support peewee."""

    class Meta:
        """Tune the handler."""

        abc: bool = True
